#include <unistd.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "ktio.h"

int kt_read_line(int     fd,
                 void  **buf)
{
    char  *cur_buf;
    char  *new_buf;
    off_t  init_pos;
    int    buf_size;
    int    bytes_read;
    int    read_size;
    int    total_read;

    /* Start by getting the current file position.  It would be nice to be
     * able to go back to where we were on error...
     */
    init_pos=lseek(fd, 0, SEEK_CUR);

    if (init_pos==-1) {
        /* Error doing the lseek... */
        return -1;
    }

    /* Set *buf == NULL; this will only be changed on successful
     * completion.
     */
    *buf=NULL;

    /* Allocate a small buffer.  buf_size will denote the size of our buffer
     * until we are done.
     */
    buf_size=KT_BUF_INC;
    total_read=0;

    cur_buf=(char *) malloc(KT_BUF_INC);

    if (cur_buf==NULL) {
        /* The malloc failed... */
        return -1;
    }

    /* Ok, now let's start reading from the file. */
    for ( ; ; ) {
        bytes_read=0;

        for ( ; ; ) {
            read_size=read(fd, cur_buf+total_read+bytes_read,
                           KT_BUF_INC-bytes_read );

            if (read_size<0) {
                if (errno==EINTR) {
                    /* Whew, the syscall just got interrupted before it could
                     * read data, that's all.
                     */
                    continue;
                }

                /* Wonderful, there was a file read error. Something
                 * stinks.
                 */
                free(cur_buf);
                lseek(fd, init_pos, SEEK_SET);
                return -1;
            }

            /* The file read was successful.  Some things _do_ work right, it
             * seems =)
             */
            bytes_read+=read_size;

            if (read_size==0) {
                /* Hmm ok then, there was apparently no more data to read.
                 * We started at EOF, it seems...
                 */
                total_read+=bytes_read;
                break;
            }

            if (bytes_read==KT_BUF_INC) {
                /* We got a full block here; we're done. */
                total_read+=bytes_read;
                break;
            }

            /* Apparently we still haven't read a full block, and there's no
             * indication that we've hit EOF.
             */
        }

        if (total_read==0) {
            /* This entire read started at EOF.  Return 0 and
             * set *buf = NULL.
             */
            free(cur_buf);
            *buf=NULL;
            return 0;
        }

        if (bytes_read==0) {
            /* It seems we reached EOF without hitting a newline
             * character.
             */
            cur_buf[total_read]='\n';
            *buf=(void *) cur_buf;
            return total_read;
        }

        new_buf=memchr((const void *) (cur_buf+total_read-bytes_read), '\n',
                       bytes_read );

        if (new_buf==NULL) {
            /* Well, there's no EOL in this block... */
            if (bytes_read<KT_BUF_INC) {
                /* Well, it seems we hit EOF without hitting a newline
                 * character.
                 */
                cur_buf[total_read]='\n';
                *buf=(void *) cur_buf;
                return total_read;
            }

            /* Ok, we may have more data to read.  Enlarge the current
             * buffer...
             */
            new_buf=(char *) malloc((total_read+KT_BUF_INC)*sizeof(char));

            if (new_buf==NULL) {
                /* Whoops, malloc failure, drop everything and run. */
                bytes_read=errno;
                free(cur_buf);
                *buf=NULL;
                lseek(fd, init_pos, SEEK_SET);
                errno=bytes_read;
                return -1;
            }

            /* Transfer from the old buffer to the new buffer... */
            memcpy(new_buf,cur_buf,total_read);
            free(cur_buf);
            cur_buf=new_buf;
            continue;
        }

        /* We seem to have hit a newline character...how big did our buffer
         * end up being?
         */
        total_read=new_buf-cur_buf;
        lseek(fd, init_pos, SEEK_SET);
        lseek(fd, total_read+1, SEEK_CUR);
        cur_buf[total_read]='\n';
        *buf=(void *) cur_buf;
        return total_read;
    }

    /* Put a little return here to make the compiler happy, even though we'll
     * never reach this point.
     */
    return -1;
}

pid_t kt_iospawnvp ( const char  **argv,
                     int          *fd_in,
                     int          *fd_out,
                     int          *fd_err )
{
    int    p_stdin[2];
    int    p_stdout[2];
    int    p_stderr[2];
    pid_t  child_id;

    /* First see if we can open the I/O pipes. */
    if (pipe(p_stdin)<0) {
        goto p_stdin_fail;
    }

    if (pipe(p_stdout)<0) {
        goto p_stdout_fail;
    }

    if (pipe(p_stderr)<0) {
        goto p_stderr_fail;
    }

    /* Ok, apparently the pipes are good to go.  See if we can fork off a
     * process...
     */
    child_id=fork();

    if (child_id<0) {
        goto fork_fail;
    }

    if (child_id==0) {
        /* We are in the child process.  Now dup to stdin, stdout, and
         * stderr.
         */
        if (dup2(p_stdin[0],0)<0) {
            goto d_stdin_fail;
        }

        if (dup2(p_stdout[1],1)<0) {
            goto d_stdout_fail;
        }

        if (dup2(p_stderr[1],2)<0) {
            goto d_stderr_fail;
        }

        /* Now close the unused pipes. */
        close(p_stdin[0]);
        close(p_stdin[1]);
        close(p_stdout[0]);
        close(p_stdout[1]);
        close(p_stderr[0]);
        close(p_stderr[1]);

        /* Now do the exec. */
        execvp((const char *) (*argv), (char * const *) argv);

        /* If we got here, then something went wrong... */
        close(0);
        close(1);
        close(2);
        exit(1);

        d_stderr_fail:
            close(1);
        d_stdout_fail:
            close(0);
        d_stdin_fail:
            close(p_stdin[0]);
            close(p_stdin[1]);
            close(p_stdout[0]);
            close(p_stdout[1]);
            close(p_stderr[0]);
            close(p_stderr[1]);
            exit(1);
    }

    /* If we got here, we're in the parent process.  Close the pipes we won't
     * be using.
     */
    close(p_stdin[0]);
    close(p_stdout[1]);
    close(p_stderr[1]);
    *fd_in=p_stdin[1];
    *fd_out=p_stdout[0];
    *fd_err=p_stderr[0];
    return child_id;

    fork_fail:
        close(p_stderr[0]);
        close(p_stderr[1]);
    p_stderr_fail:
        close(p_stdout[0]);
        close(p_stdout[1]);
    p_stdout_fail:
        close(p_stdin[0]);
        close(p_stdin[1]);
    p_stdin_fail:
        return -1;
}

pid_t kt_spawnvp(char **argv)
{
    pid_t child_pid;

    child_pid=fork();

    if (child_pid<0) {
        /* The fork failed! */
        return -1;
    }

    if (child_pid==0) {
        /* We are in the child process. */
        execvp(*argv, argv);

        /* If we're still here, we got an error! */
        exit(1);
    }

    /* We're in the parent process. */
    return child_pid;
}

pid_t kt_spawnsys(const char *syscmd)
{
    pid_t ret;
    int   result;

    ret=fork();

    if (ret<0) {
        /* The fork failed! */
        return -1;
    }

    if (ret==0) {
        /* We are in the child process. */
        result=system(syscmd);

        if (result==-1) {
            /* The system() command apparently failed... */
            exit(result);
        }

        exit(WEXITSTATUS(result));
    }

    /* We are in the parent process. */
    return ret;
}
