#ifndef __KTIO_H__
#  define __KTIO_H__

#  define KT_BUF_INC    512

#  include <unistd.h>

#  ifdef __cplusplus

extern "C" {

#  endif

int kt_read_line ( int fd, void **buf );
/* Added 3/11/2001 by Kelledin

    DESCRIPTION:
        Reads a line from the file structure referenced by fd and stores it in
        a dynamically allocated buffer.

    PRECONDITIONS:
        buf must point to a valid pointer (*buf can be anything; it will be
        overwritten by kt_read_line()).

    POSTCONDITIONS:
        *buf will point to a buffer containing the line read from the file,
        or NULL if no line could be read.  If a newline character was
        encountered, then it is read as part of the line.  If EOF is
        encountered with no newline, then an extra '\n' character will be
        tacked on to the end of *buf.  If EOF was encountered before any
        data was read, or an error occurs, then *buf is set to NULL.

    RETURNS:
        On success, kt_read_line() returns the number of bytes read from the
        file, not including any terminating '\n' character it encounters.  On
        failure, kt_read_line() returns -1 and sets errno to indicate what
        croaked.

    REMARKS:
        kt_read_line() does not recognize EOF as an error condition; if it is
        asked to read from a file descriptor whose pointer is already
        positioned at the end of the file, it will simply set *buf == NULL
        and return 0.  Thus if ( kt_read_line ( fd, &ptr ) == 0 && *buf ==
        NULL ), then the file pointer was already at EOF.  This makes a nice
        terminating case in loops.

        On failure, kt_read_line() will do its best to set everything back the
        way it was, except for overwriting the value of *buf.  This includes
        setting the file pointer back to its initial value.

        kt_read_line() can read lines of indefinite length; if necessary, it
        will repeatedly enlarge its read buffer until it runs out of available
        memory.  However, it will never leave more than a certain amount of
        memory unused in the buffer; this certain amount is currently defined
        as KT_BUF_INC in ktio.h and is currently equal to 512.  DO NOT expect
        it to always be at this value!

        Remember to call free() on *buf when you're done with it.

*/

pid_t kt_iospawnvp ( const char **argv, int *fd_in, int *fd_out, int *fd_err );
/* Added 3/12/2001 by Kelledin

    DESCRIPTION:
        Spawns off a child process of a given name and connects its standard
        input, output, and error streams to file descriptors.  The parent
        process continues to run asynchronously and does not wait for the child
        to complete.  argv is synonymous to the second parameter of execvp();
        the name of the process to spawn is the first element of argv.  The
        name of the process is not required as a parameter by itself because
        it is already in argv; I consider execvp's way of doing it to be
        unnecessary redundancy.

    PRECONDITIONS:
        All parameters must point to valid buffers.

    POSTCONDITIONS:
        A new child process will exist, and *fd_in, *fd_out, and *fd_err will
        be pipe descriptors connected to the child's stdin, stdout, and stderr
        respectively.  *argv will remain unchanged.

    RETURNS:
        Returns the process ID of the new child on success; otherwise, returns a
        negative value and leaves *fd_in, *fd_out, and *fd_err unchanged.

    REMARKS:
        This is handy if you want to inspect a child's standard I/O streams,
        send the child predefined input, or just suppress a lot of ugly
        messages.  Remember to close the descriptors when you're done, but
        also remember that closing one end of a pipe prematurely may generate
        an ugly error message on the console and a broken process.

*/

pid_t kt_spawnvp ( char **argv );
/* Added 3/13/2001 by Kelledin

    DESCRIPTION:
        Identical to kt_iospawnvp, except that the child's stdin, stdout, and
        stderr are left at their default (usually the system console).

    PRECONDITIONS:
        argv must point to a valid buffer.

    POSTCONDITIONS:
        A new child process will exist.

    RETURNS:
        Returns the process ID of the new child on success; otherwise, returns
        a negative value.

    REMARKS:
        Yes, this should have been a predecessor to kt_iospawnvp().  I did
        it a little out-of-order because I had a more pressing need for the
        kt_iospawnvp() functionality.  kt_spawnvp() would be superceded by
        kt_iospawnvp(), except that kt_spawnvp() is lighter, faster, and less
        likely to fail because it doesn't take the time or resources to set up
        all the standard I/O interconnections.

*/

pid_t kt_spawnsys(const char *syscmd);

#  ifdef __cplusplus

}

#  endif

#endif
