/***************************************************************************
         misc.c  -  miscellaneous helper functions for various tasks.
                             -------------------
    begin                : Wed Dec 12 2001
    copyright            : (C) 2001 by Kelledin
    email                : kelledin@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of version 2 of the GNU General Public License as  *
 *   published by the Free Software Foundation.                            *
 *                                                                         *
 ***************************************************************************/

#include "mondo.h"

char         *conf_file      = NULL;
int           mondo_delay    = 0;
int           pid_fd         = -1;
int           dolock         = 1;
struct flock  pid_lock;

int lock_pidfile()
{
    char  pid_str[20];
    int   result;
    int   bytes_written;
    int   total_written;

    if (pid_fd>=0) {
        /* lock_pidfile already called... */
        syslog(LOG_ERR, "lock_pidfile(): already called!");
        errno=EAGAIN;
        return -1;
    }

    result=snprintf(pid_str, 18, "%d", getpid());

    if (result<0 || result>=18) {
        syslog(LOG_ERR, "converting int to char *: %s",
               strerror(errno));
        return -1;
    }

    pid_fd=open("/var/run/mondo.pid", O_CREAT|O_TRUNC|O_WRONLY,
                S_IRUSR|S_IWUSR);

    if (pid_fd<0) {
        syslog(LOG_ERR, "Opening /var/run/mondo.pid: %s",
                strerror(errno));

        return -1;
    }

    /* Now lock the pidfile. */
    pid_lock.l_type=F_WRLCK;
    pid_lock.l_whence=SEEK_SET;
    pid_lock.l_start=0;
    pid_lock.l_len=0;
    pid_lock.l_pid=getpid();

    if (fcntl(pid_fd, F_SETLK, &pid_lock)<0) {
        if (errno==EACCES || errno==EAGAIN) {
            fcntl(pid_fd, F_GETLK, &pid_lock);
            syslog(LOG_ERR, "It seems I'm already running as PID %d!",
                   pid_lock.l_pid);
            syslog(LOG_ERR, "Exiting now.");
        }
        else {
            syslog(LOG_ERR, "Locking /var/run/mondo.pid: %s",
                   strerror(errno));
        }

        return -1;
    }

    /* Now write the PID. */
    for (total_written=0; total_written<result; total_written+=bytes_written) {
        bytes_written=write(pid_fd, pid_str+total_written,
                            result-total_written);

        if (bytes_written<0) {
            if (errno!=EINTR) {
                syslog(LOG_ERR, "Writing /var/run/mondo.pid: %s",
                       strerror(errno));

                pid_lock.l_type=F_UNLCK;
                pid_lock.l_whence=SEEK_SET;
                pid_lock.l_start=0;
                pid_lock.l_len=0;
                pid_lock.l_pid=getpid();
                fcntl(pid_fd, F_SETLK, &pid_lock);
                close(pid_fd);
                return -1;
            }

            bytes_written=0;
        }
    }

    /* Ok, we're finished here... */
    return 0;
}

int unlock_pidfile()
{
    if (pid_fd<0) {
        syslog(LOG_ERR, "unlock_pidfile(): /var/run/mondo.pid is not open!");
        errno = EPERM;
        return -1;
    }

    /* Now unlock the pidfile. */
    pid_lock.l_type=F_UNLCK;
    pid_lock.l_whence=SEEK_SET;
    pid_lock.l_start=0;
    pid_lock.l_len=0;
    pid_lock.l_pid=getpid();

    if (fcntl(pid_fd, F_SETLK, &pid_lock)<0) {
        syslog(LOG_ERR, "unlocking /var/run/mondo.pid: %s\n",
               strerror(errno));

        return -1;
    }

    /* Now close and remove the file. */
    close(pid_fd);
    unlink("/var/run/mondo.pid");
    return 0;
}

void sighup_handler(int sig)
{
    free_sensor_list();
    syslog(LOG_INFO, "Received SIGHUP.");

    if (read_conf()<0) {
        syslog(LOG_ERR, "ERROR: Reading %s: %s", conf_file, strerror(errno));
        syslog(LOG_ERR, "Exiting now." );
        shutdown_lm_api();

        if (dolock) {
            unlock_pidfile();
        }

        exit(1);
    }

    if (set_alrm_timer(mondo_delay)<0) {
        syslog(LOG_ERR, "ERROR: Setting alarm timer: %s", strerror(errno));
    }
}

/*void sigchld_handler(int sig)
{
    int status;

    syslog(LOG_INFO, "Received SIGCHLD");

    if (wait(&status)<0) { syslog(LOG_INFO, "lost SIGCHLD: %s", strerror(errno)); }
}*/

void sigterm_handler(int sig)
{
    free_sensor_list();
    syslog(LOG_INFO, "Received SIGTERM, exiting now.");

    if (dolock) {
        unlock_pidfile();
    }

    shutdown_lm_api();

    if (conf_file!=NULL) {
        free(conf_file);
    }

    exit(0);
}

void sigint_handler(int sig)
{
    free_sensor_list();
    syslog(LOG_INFO, "Received SIGINT, exiting now.");

    if (dolock) {
        unlock_pidfile();
    }

    shutdown_lm_api();

    if (conf_file!=NULL) {
        free(conf_file);
    }

    exit(0);
}

void sigsegv_handler(int sig)
{
    syslog(LOG_ERR, "Sorry, I have just blown a very large hole in my head.");
    syslog(LOG_ERR, "Send a bug report to kelledin@users.sourceforge.net.");

    if (dolock) {
        unlock_pidfile();
    }

    if (conf_file!=NULL) {
        free(conf_file);
    }

    exit(1);
}

int set_alrm_timer(int msecs)
{
    struct itimerval ltimer;

    ltimer.it_interval.tv_sec=msecs/1000;
    ltimer.it_interval.tv_usec=(msecs%1000)*1000;
    ltimer.it_value=ltimer.it_interval;
    return (setitimer(ITIMER_REAL, &ltimer, NULL));
}

int set_signal_handlers()
{
    struct sigaction new_action;

    new_action.sa_handler=sigterm_handler;
    new_action.sa_flags=SA_RESTART;

    if (sigfillset(&(new_action.sa_mask))<0
     || sigaction(SIGTERM, &new_action, NULL)<0) {
        return -1;
    }

    new_action.sa_handler=SIG_IGN;
    new_action.sa_flags=SA_NOCLDSTOP|SA_RESTART;

    if (sigaction(SIGCHLD, &new_action, NULL)<0) {
        return -1;
    }

    new_action.sa_handler=sighup_handler;
    new_action.sa_flags=SA_RESTART;

    if (sigfillset(&(new_action.sa_mask))<0
     || sigaction(SIGHUP, &new_action, NULL)<0) {
        return -1;
    }

    new_action.sa_handler=sigalrm_handler;

    if (sigdelset(&(new_action.sa_mask), SIGTERM)<0
     || sigdelset(&(new_action.sa_mask), SIGINT)<0
     || sigaction(SIGALRM, &new_action, NULL)<0) {
        return -1;
    }

    new_action.sa_handler=sigsegv_handler;

    if (sigfillset(&(new_action.sa_mask))<0
     || sigaction(SIGSEGV, &new_action, NULL)<0) {
        return -1;
    }

    new_action.sa_handler=sigint_handler;

    if (sigaction(SIGINT, &new_action, NULL)<0) {
        return -1;
    }

    return 0;
}

char **parse_line(const char *line, char *delim, int *argc)
{
    char       **ret;
    int          nargs;
    int          arg_size;
    const char  *tmp;
    char        *ctmp;

    if (strchr(delim, '"')!=NULL) {
        /* The quote character cannot be used as an identifier! */
        errno=EINVAL;
        return NULL;
    }

    tmp=line;

    /* Scope out the beginning if the line. */
    tmp+=strspn(tmp, delim);

    for (nargs=0; *tmp!='\0'; nargs++) {
        /* Does this parameter begin with a quote? */
        if (*tmp=='"') {
            /* It does begin with a quote.  Seek to the next quote... */
            for (tmp++; *tmp != '"'; ) {
                if (*tmp=='\0') {
                    /* Unterminated quote... */
                    errno=EINVAL;
                    return NULL;
                }

                if (*tmp=='\\') {
                    /* Escaped character. */
                    tmp++;

                    if (*tmp=='\0') {
                        /* Unterminated quote... */
                        errno=EINVAL;
                        return NULL;
                    }
                }

                tmp++;
            }

            /* tmp now points to the ending quote of the current arg. */
            tmp++;

            if (*tmp!='\0' && strchr(delim, *tmp)==NULL) {
                /* The next character is not a delimiter or a terminator!
                 * This is bad...
                 */
                errno=EINVAL;
                return NULL;
            }
        }
        else {
            /* The line does not begin with a quote.  Seek to the first
             * delimiting character...
             */
            for ( ; *tmp!='\0' && strchr(delim, *tmp)==NULL; ) {
                if (*tmp=='"') {
                    errno=EINVAL;
                    return NULL;
                }

                if (*tmp=='\\') {
                    /* It's an escape character. */
                    tmp++;

                    if (*tmp=='\0') {
                        /* Incomplete escape sequence. */
                        errno=EINVAL;
                        return NULL;
                    }
                }

                tmp++;
            }
        }

        /* Ok, we're to the point just past the current arg.  Seek to the
         * first non-delimiting character...
         */
        tmp+=strspn(tmp, delim);
    }

    /* Ok, we now have the number of arguments. */
    ret=(char **) malloc((nargs+1)*sizeof(char *));

    if (ret==NULL) {
        return NULL;
    }

    ret[nargs]=NULL;
    *argc=nargs;

    /* Now extract each argument... */
    for (nargs=0; nargs<*argc; nargs++) {
        /* Seek to the first non-delimiting character. */
        line+=strspn(line, delim);

        if (*line=='"') {
            /* This argument is a quoted argument. */
            line++;
            tmp=line;

            for (arg_size=1; *tmp!='"'; arg_size++) {
                if (*tmp=='\\') {
                    /* Escaped character. */
                    tmp+=2;
                }
                else {
                    tmp++;
                }
            }

            ctmp=(char *) malloc(arg_size);

            if (ctmp==NULL) {
                for (nargs--; nargs>=0; nargs--) {
                    free(ret[nargs]);
                }

                free(ret);
                return NULL;
            }

            /* Now copy over the data. */
            for (ret[nargs]=ctmp; *line!='"'; ctmp++) {
                if (*line=='\\') {
                    /* It's an escape character. */
                    line++;
                }

                *ctmp=*line;
                line++;
            }

            tmp++;
        }
        else {
            /* It's not a quoted parameter... */
            tmp=line;

            for (arg_size=1;
                 *tmp!='\0' && strchr(delim, *tmp)==NULL;
                 arg_size++ ) {
                if (*tmp=='\\') {
                    /* It's an escape character. */
                    tmp+=2;
                }
                else {
                    tmp++;
                }
            }

            ctmp=(char *) malloc(arg_size);

            if (ctmp==NULL) {
                for (nargs--; nargs>=0; nargs--) {
                    free(ret[nargs]);
                }

                free(ret);
                return NULL;
            }

            /* Now copy over the data. */
            for (ret[nargs]=ctmp;
                 *line!='\0' && strchr(delim, *line)==NULL;
                 ctmp++ ) {
                if (*line=='\\') {
                    /* It's an escape character. */
                    line++;
                }

                *ctmp=*line;
                line++;
            }
        }

        *ctmp='\0';
        line=tmp;
    }

    return ret;
}

int read_conf()
{
    int            fd;
    int            argc;
    int            bytes_read;
    int            lnum;
    double         tlimit;
    char          *ctmp;
    char          *lbuf;
    char         **argv;
    char          *tcmd;
    mondo_sensor  *lsens;
    int            mondo_log;
    int            mondo_throttle;
    int            mondo_grace;

    fd=open(conf_file, O_RDONLY);

    if (fd<0) {
        fprintf(stderr, "FATAL: opening %s: %s\n", conf_file, strerror(errno));
        return -1;
    }

    mondo_delay=mondo_grace=0;
    mondo_log=mondo_throttle=-1;

    for (lnum=1; ; lnum++) {
        bytes_read=kt_read_line(fd, (void **) &lbuf);

        if (bytes_read<0) {
            /* Error */
            fprintf(stderr, "FATAL: reading %s: %s\n", conf_file,
                    strerror(errno));
            goto fail_list_started;
        }

        if (lbuf==NULL) {
            /* EOF */
            break;
        }

        if (*lbuf=='#') {
            /* Comment */
            free(lbuf);
            continue;
        }

        lbuf[bytes_read]='\0';

        if (strspn(lbuf, "\t ")==bytes_read) {
            /* Blank line. */
            free(lbuf);
            continue;
        }

        /* Now attempt to parse the line... */
        argv=parse_line(lbuf, "\t ", &argc);
        free(lbuf);

        if (argv==NULL) {
            /* Error parsing the line... */
            fprintf(stderr, "FATAL: parsing line %d: %s\n", lnum,
                    strerror(errno));

            goto fail_list_started;
        }

        if (argc>2) {
            fprintf(stderr, "FATAL: parsing line %d: too many parameters!\n",
                    lnum);

            goto fail_argv_alloced;
        }

        if (argc<2) {
            fprintf(stderr, "FATAL: parsing line %d: not enough parameters!\n",
                    lnum );

            goto fail_argv_alloced;
        }

        if (strcasecmp(*argv, "DELAY")==0) {
            /* It's a DELAY option. */
            if (mondo_delay!=0) {
                fprintf(stderr,
                        "FATAL: parsing line %d: 'DELAY' already specified!\n",
                        lnum);

                goto fail_argv_alloced;
            }

            mondo_delay=strtol(argv[1], &ctmp, 0);

            if (*ctmp!='\0' || mondo_delay<=0) {
                fprintf(stderr, "FATAL: parsing line %d: invalid parameter.\n",
                        lnum);

                goto fail_argv_alloced;
            }

            for (argc--; argc>=0; argc--) {
                free(argv[argc]);
            }

            free(argv);
        }
        else if (strcasecmp(*argv, "LOG")==0) {
            /* it's a LOG option. */
            if (mondo_log>=0) {
                fprintf(stderr,
                        "FATAL: parsing line %d: 'LOG' already specified!\n",
                        lnum);

                goto fail_argv_alloced;
            }

            if (strcasecmp(argv[1], "YES")==0
             || strcasecmp(argv[1], "TRUE")==0
             || strcmp(argv[1], "1")==0) {
                mondo_log=1;
            }
            else if (strcasecmp(argv[1], "NO")==0
                  || strcasecmp(argv[1], "FALSE")==0
                  || strcasecmp(argv[1], "0")==0) {
                mondo_log=0;
            }
            else {
                fprintf(stderr,
                        "FATAL: parsing line %d: invalid parameter!\n",
                        lnum);

                goto fail_argv_alloced;
            }
        }
        else if (strcasecmp(*argv, "THROTTLE")==0) {
            /* it's a THROTTLE option. */
            if (mondo_throttle>=0) {
                fprintf(stderr,
                        "FATAL: parsing line %d: 'THROTTLE' already specified!\n",
                        lnum);

                goto fail_argv_alloced;
            }

            mondo_throttle=strtol(argv[1], &ctmp, 0);

            if (*ctmp!='\0' || mondo_throttle<=0) {
                fprintf(stderr, "FATAL: parsing line %d: invalid parameter.\n",
                        lnum);

                goto fail_argv_alloced;
            }

            for (argc--; argc>=0; argc--) {
                free(argv[argc]);
            }

            free(argv);
        }
        else if (strcasecmp(*argv, "GRACE")==0) {
            /* it's a grace option. */
            if (mondo_grace>0) {
                fprintf(stderr,
                        "FATAL: parsing line %d: 'GRACE' already specified!\n",
                        lnum);

                goto fail_argv_alloced;
            }

            mondo_grace=strtol(argv[1], &ctmp, 0);

            if (*ctmp!='\0' || mondo_grace<=0) {
                fprintf(stderr, "FATAL: parsing line %d: invalid parameter.\n",
                        lnum);

                goto fail_argv_alloced;
            }

            for (argc--; argc>=0; argc--) {
                free(argv[argc]);
            }

            free(argv);
        }
        else if (strcmp(argv[1], "{")==0) {
            /* It's a specifier for a certain sensor's settings. */
            lsens=locate_sensor(argv[0]);

            if (lsens==NULL) {
                fprintf(stderr, "FATAL: locating sensor feature %s: %s\n",
                        argv[0], strerror(errno));

                free(argv[0]);
                free(argv[1]);
                free(argv);
                goto fail_list_started;
            }

            free(argv[0]);
            free(argv[1]);
            free(argv);

            for (lnum++; ; lnum++) {
                bytes_read=kt_read_line(fd, (void **) &lbuf);

                if (bytes_read<0) {
                    /* Error */
                    fprintf(stderr, "FATAL: reading %s: %s\n", conf_file,
                            strerror(errno));

                    goto fail_list_started;
                }

                if (lbuf==NULL) {
                    /* Premature EOF! */
                    fprintf(stderr, "FATAL: line %d: missing '}' before EOF.n",
                            lnum);

                    goto fail_list_started;
                }

                if (*lbuf=='#') {
                    /* Comment */
                    free(lbuf);
                    continue;
                }

                lbuf[bytes_read]='\0';

                if (strspn(lbuf, "\t ")==bytes_read) {
                    /* Blank line. */
                    free(lbuf);
                    continue;
                }

                /* Now attempt to parse the line... */
                argv=parse_line(lbuf, "\t ", &argc);
                free(lbuf);

                if (argv==NULL) {
                    /* Error parsing the line... */
                    fprintf(stderr, "FATAL: line %d: %s\n", lnum,
                            strerror(errno));

                    goto fail_list_started;
                }

                if (strcmp(*argv, "}")==0) {
                    /* This should be a closing brace. */
                    if (argc!=1) {
                        fprintf(stderr,
                                "FATAL: line %d: excess parameters following '}'\n",
                                lnum);

                        goto fail_argv_alloced;
                    }

                    for (argc--; argc>=0; argc--) {
                        free(argv[argc]);
                    }

                    free(argv);
                    break;
                }

                if (strcasecmp(*argv, "ABOVE")==0) {
                    /* It's an upper bound directive. */
                    if (argc>3) {
                        /* Too many parameters! */
                        fprintf(stderr,
                                "FATAL: line %d: too many parameters!\n",
                                lnum);

                        goto fail_argv_alloced;
                    }

                    if (argc<3) {
                        /* Not enough parameters! */
                        fprintf(stderr,
                                "FATAL: line %d: not enough parameters!\n",
                                lnum);

                        goto fail_argv_alloced;
                    }

                    tlimit=strtod(argv[1], &ctmp);

                    if (*ctmp!='\0') {
                        fprintf(stderr, "FATAL: line %d: %s\n", lnum,
                                strerror(errno));

                        goto fail_argv_alloced;
                    }

                    tcmd=strdup(argv[2]);

                    if (tcmd==NULL) {
                        fprintf(stderr, "FATAL: line %d: %s\n", lnum,
                                strerror(errno));

                        goto fail_argv_alloced;
                    }

                    if (add_ubound(&(lsens->ubound), tlimit, tcmd)<0) {
                        fprintf(stderr, "FATAL: line %d: %s\n", lnum,
                                strerror(errno));

                        free(tcmd);
                        goto fail_argv_alloced;
                    }
                }
                else if (strcasecmp(*argv, "BELOW")==0) {
                    /* It's a lower bound directive. */
                    if (argc>3) {
                        /* Too many parameters! */
                        fprintf(stderr,
                                "FATAL: line %d: too many parameters!\n",
                                lnum);

                        goto fail_argv_alloced;
                    }

                    if (argc<3) {
                        /* Not enough parameters! */
                        fprintf(stderr,
                                "FATAL: line %d: not enough parameters!\n",
                                lnum);

                        goto fail_argv_alloced;
                    }

                    tlimit=strtod(argv[1], &ctmp);

                    if (*ctmp!='\0') {
                        fprintf(stderr, "FATAL: line %d: %s\n", lnum,
                                strerror(errno));

                        goto fail_argv_alloced;
                    }

                    tcmd=strdup(argv[2]);

                    if (tcmd==NULL) {
                        fprintf(stderr, "FATAL: line %d: %s\n", lnum,
                                strerror(errno));

                        goto fail_argv_alloced;
                    }

                    if (add_lbound(&(lsens->lbound), tlimit, tcmd)<0) {
                        fprintf(stderr, "FATAL: line %d: %s\n", lnum,
                                strerror(errno));

                        free(tcmd);
                        goto fail_argv_alloced;
                    }
                }
                else if (strcasecmp(*argv, "THROTTLE")==0) {
                    /* It's a throttle directive. */
                    if (argc>2) {
                        /* Too many parameters! */
                        fprintf(stderr,
                                "FATAL: line %d: too many parameters!\n",
                                lnum);

                        goto fail_argv_alloced;
                    }

                    if (argc<2) {
                        /* Not enough parameters! */
                        fprintf(stderr,
                                "FATAL: line %d: not enough parameters!\n",
                                lnum);

                        goto fail_argv_alloced;
                    }

                    if (lsens->throttle>=0) {
                        fprintf(stderr,
                                "FATAL: line %d: throttle level aready specified!\n",
                                lnum);

                        goto fail_argv_alloced;
                    }

                    lsens->throttle=strtol(argv[1], &ctmp, 0);

                    if (*ctmp!='\0') {
                        fprintf(stderr, "FATAL: line %d: %s\n", lnum,
                                strerror(errno));

                        goto fail_argv_alloced;
                    }
                    else if (lsens->throttle<0) {
                        fprintf(stderr,
                                "FATAL: line %d: invalid throttle value!\n",
                                lnum);

                        goto fail_argv_alloced;
                    }
                }
                else if (strcasecmp(*argv, "GRACE")==0) {
                    /* It's a grace directive. */
                    if (argc>2) {
                        /* Too many parameters! */
                        fprintf(stderr,
                                "FATAL: line %d: too many parameters!\n",
                                lnum);

                        goto fail_argv_alloced;
                    }

                    if (argc<2) {
                        /* Not enough parameters! */
                        fprintf(stderr,
                                "FATAL: line %d: not enough parameters!\n",
                                lnum);

                        goto fail_argv_alloced;
                    }

                    if (lsens->grace>0) {
                        fprintf(stderr,
                                "FATAL: line %d: grace period aready specified!\n",
                                lnum);

                        goto fail_argv_alloced;
                    }

                    lsens->grace_ticks=lsens->grace=strtol(argv[1], &ctmp, 0);

                    if (*ctmp!='\0') {
                        fprintf(stderr, "FATAL: line %d: %s\n", lnum,
                                strerror(errno));

                        goto fail_argv_alloced;
                    }
                    else if (lsens->grace<=0) {
                        fprintf(stderr,
                                "FATAL: line %d: invalid grace period!\n",
                                lnum);

                        goto fail_argv_alloced;
                    }
                }
                else if (strcasecmp(*argv, "LOG")==0) {
                    if (lsens->log>=0) {
                        fprintf(stderr,
                                "FATAL: line %d: 'LOG' already specified!\n",
                                lnum);

                        goto fail_argv_alloced;
                    }

                    if (strcasecmp(argv[1], "YES")==0
                     || strcasecmp(argv[1], "TRUE")==0
                     || strcmp(argv[1], "1")==0) {
                        lsens->log=1;
                    }
                    else if (strcasecmp(argv[1], "NO")==0
                          || strcasecmp(argv[1], "FALSE")==0
                          || strcasecmp(argv[1], "0")==0) {
                        lsens->log=0;
                    }
                    else {
                        fprintf(stderr,
                                "FATAL: parsing line %d: invalid parameter!\n",
                                lnum);

                        goto fail_argv_alloced;
                    }
                }
                else {
                    fprintf(stderr, "FATAL: line %d: unknown directive '%s'\n",
                            lnum, *argv);

                    goto fail_argv_alloced;
                }

                for (argc--; argc>=0; argc--) {
                    free(argv[argc]);
                }

                free(argv);
            }

            lnum++;
        }
        else {
            fprintf(stderr, "FATAL: line %d: unknown directive '%s'\n", lnum,
                    *argv );

            goto fail_argv_alloced;
        }
    }

    close(fd);

    if (mondo_log<0) {
        mondo_log=0;
    }

    if (mondo_throttle<0) {
        mondo_throttle=0;
    }

    if (mondo_delay==0) {
        mondo_delay=4000;
    }

    set_log_option(mondo_log);
    set_throttle_option(mondo_throttle);
    set_grace_option(mondo_grace);
    return 0;

    fail_argv_alloced:
        for (argc--; argc>=0; argc--) {
            free(argv[argc]);
        }

        free(argv);

    fail_list_started:
        free_sensor_list();
        close(fd);
        return -1;
}

int add_ubound(mondo_action **ulist,
               double         lim,
               char          *cmd)
{
    mondo_action *tmp;
    mondo_action *nact;

    if (ulist==NULL || cmd==NULL) {
        errno=EINVAL;
        return -1;
    }

    nact=(mondo_action *) malloc(sizeof(mondo_action));

    if (nact==NULL) {
        errno=ENOMEM;
        return -1;
    }

    nact->limit=lim;
    nact->syscmd=cmd;

    if (*ulist==NULL || (*ulist)->limit>lim) {
        nact->next=*ulist;
        *ulist=nact;
        return 0;
    }

    tmp=*ulist;

    while (tmp->next!=NULL && tmp->next->limit<=lim) {
        tmp=tmp->next;
    }

    nact->next=tmp->next;
    tmp->next=nact;
    return 0;
}

int add_lbound(mondo_action **llist,
               double         lim,
               char          *cmd)
{
    mondo_action *tmp;
    mondo_action *nact;

    if (llist==NULL || cmd==NULL) {
        errno=EINVAL;
        return -1;
    }

    nact=(mondo_action *) malloc(sizeof(mondo_action));

    if (nact==NULL) {
        errno=ENOMEM;
        return -1;
    }

    nact->limit=lim;
    nact->syscmd=cmd;

    if (*llist==NULL || (*llist)->limit<lim) {
        nact->next=*llist;
        *llist=nact;
        return 0;
    }

    tmp=*llist;

    while (tmp->next!=NULL && tmp->next->limit>=lim) {
        tmp=tmp->next;
    }

    nact->next=tmp->next;
    tmp->next=nact;
    return 0;
}

void free_alist(mondo_action *alist)
{
    mondo_action *tmp  = NULL;
    mondo_action *prev = NULL;

    for (tmp=alist; tmp!=NULL; tmp=tmp->next) {
        if (prev!=NULL) {
            free(prev->syscmd);
            free(prev);
        }

        prev=tmp;
    }

    if (prev!=NULL) {
        free(prev->syscmd);
        free(prev);
    }
}
