/***************************************************************************
         mondo-setup.c  -  source file for a rudimentary setup utility
                             -------------------
    begin                : Tue Dec 18 07:05:40 CST 2001
    copyright            : (C) 2001 by Kelledin
    email                : kelledin@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of version 2 of the GNU General Public License as  *
 *   published by the Free Software Foundation.                            *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <sensors/sensors.h> /* The lm_sensors API. */

#ifndef SENSORS_CONF_FILE
#  define SENSORS_CONF_FILE "/etc/sensors.conf"
#endif

#ifndef MONDO_VERSION
#  define MONDO_VERSION "0.1"
#endif

/* Define a maximum scanning time of 20160 minutes (2 weeks). */
#define MAX_SCAN_MINUTES    20160

typedef struct __tag_mondo_sensor {

    const sensors_chip_name *chip_id;
    char *feature_label;
    const char *bus_type;
    int feature_id;
    int i2c_num;
    double upper;
    double lower;
    struct __tag_mondo_sensor *next;

} mondo_sensor;

/* Variables for the global linked list of sensor features. */
mondo_sensor *list_head = NULL;
mondo_sensor *list_tail = NULL;

mondo_sensor *new_sensor(const sensors_chip_name *chip_id, const char *label, int i2c_nr, int feature);

void read_sensor(mondo_sensor *sens);

int write_header(FILE *fd);

void usage(const char *cmd, FILE *fd);

int main(int argc, char **argv)
{
    mondo_sensor               *lsens;
    int                         opt_char;
    int                         secs;
    int                         mins;
    int                         hours;
    int                         mod;
    int                         cur_msecs;
    int                         i2c_idx;
    int                         grace_period = 0;
    int                         force        = 0;
    int                         result       = 0;
    int                         delay        = 0;
    int                         minutes      = 0;
    int                         total_msecs  = 0;
    int                         chip_idx     = 0;
    int                         throttle     = -1;
    int                         log_opt      = 0;
    int                         i2c_nr       = 0;
    struct stat                 file_stat;
    const sensors_chip_name    *chip_name;
    const char                 *conf_file    = NULL;
    const char                 *sensors_file = NULL;
    const char                 *opt_string = "o:d:t:s:T:g:fhlV";
    char                       *ctmp;
    FILE                       *conf_fd;
    FILE                       *sensors_fd;

    /* First, gather command-line options. */
    do {
        opt_char=getopt(argc, argv, opt_string);

        switch(opt_char) {
        case -1:
            break;

        case 'o':
            if (conf_file!=NULL) {
                fprintf(stderr, "FATAL: '-o' option specified twice!\n");
                return -1;
            }

            conf_file=optarg;
            break;

        case 'd':
            if (delay!=0) {
                fprintf(stderr, "FATAL: '-d' option specified twice!\n");
                return -1;
            }

            delay=strtol(optarg, &ctmp, 0);

            if (*optarg=='\0' || *ctmp!='\0' || delay<=0) {
                fprintf(stderr, "FATAL: invalid delay interval!\n");
                return -1;
            }

            break;

        case 't':
            if (minutes!=0) {
                fprintf(stderr, "FATAL: '-t' option specified twice!\n");
                return -1;
            }

            minutes=strtol(optarg, &ctmp, 0);

            if (*optarg=='\0'
             || *ctmp!='\0'
             || minutes<=0
             || minutes>MAX_SCAN_MINUTES) {
                fprintf(stderr, "FATAL: invalid running time!\n");
                return -1;
            }

            break;

        case 'g':
            if (grace_period!=0) {
                fprintf(stderr, "FATAL: '-g' option specified twice!\n");
                return -1;
            }

            grace_period=strtol(optarg, &ctmp, 0);

            if (*optarg=='\0'
             || *ctmp!='\0'
             || grace_period<=0
             || minutes>MAX_SCAN_MINUTES) {
                fprintf(stderr, "FATAL: invalid grace period!\n");
                return -1;
            }

            break;

        case 'T':
            if (throttle>=0) {
                fprintf(stderr, "FATAL: '-T' option specified twice!\n");
                return -1;
            }

            throttle=strtol(optarg, &ctmp, 0);

            if (*optarg=='\0'
             || *ctmp!='\0'
             || throttle<=0) {
                fprintf(stderr, "FATAL: invalid throttle setting!\n");
                return -1;
            }

            break;

        case 's':
            if (sensors_file!=NULL) {
                fprintf(stderr, "FATAL: '-s' option specified twice!\n");
                return -1;
            }

            sensors_file=optarg;
            break;

        case 'f':
            force=1;
            break;

        case 'h':
            usage(*argv, stdout);
            return 0;

        case 'l':
            log_opt=1;
            break;

        case 'V':
            printf("System Monitor setup v%s\n\n", MONDO_VERSION);
            return 0;

        case '?':
        default:
            usage(*argv, stderr);
            return -1;
        }
    } while (opt_char!=-1);

    if (conf_file==NULL) {
        /* Output file not specified; use the default. */
        conf_file="./mondo.conf";
    }

    if (minutes==0) {
        /* Scan time not specified; use the default. */
        minutes=120;
    }

    if (sensors_file==NULL) {
        /* sensors.conf file not specified; use the default. */
        sensors_file=SENSORS_CONF_FILE;
    }

    if (delay==0) {
        /* Polling interval not specified; use the default. */
        delay=4000;
    }

    if (force==0) {
        /* Make sure the new config file doesn't already exist. */
        if (stat(conf_file, &file_stat)==0) {
            /* D'oh, yes it does! */
            fprintf(stderr, "FATAL: %s already exists!\n", conf_file);
            return -1;
        }
    }

    /* Open the sensors.conf file for reading. */
    sensors_fd=fopen(sensors_file, "r");

    if (sensors_fd==NULL) {
        fprintf(stderr, "FATAL: opening %s: %s\n", sensors_file,
                strerror(errno));
        return -1;
    }

    /* Create/open the new config file for writing. */
    conf_fd=fopen(conf_file, "w");

    if (conf_fd==NULL) {
        fprintf(stderr, "FATAL: opening %s: %s\n", conf_file, strerror(errno));
        goto fail_sensors_open;
    }

    printf("Scanning sensors -- basic configuration will be written to %s\n", conf_file);

    if (minutes>1) {
        printf("I will scan for approximately %d minutes before writing this file.\n", minutes);
    }
    else {
        printf("I will scan for approximately 1 minute before writing this file.\n");
    }

    if (sensors_init(sensors_fd)!=0) {
        fprintf(stderr, "FATAL: initializing sensors: %s\n", strerror(errno));
        goto fail_conf_open;
    }

    /* Go into the sensor chip detection loop. */
    for (chip_name=sensors_get_detected_chips(&chip_idx);
         chip_name!=NULL;
         chip_name=sensors_get_detected_chips(&chip_idx)) {
        char                       *bus_type;
        int                         fidx1;
        int                         fidx2;
        const sensors_feature_data *feature_id;

        switch(chip_name->bus) {
        case SENSORS_CHIP_NAME_BUS_ISA:
            bus_type="isa";
            printf("Found chip: %s-%s-%x\n", chip_name->prefix, bus_type,
                   chip_name->addr);
            i2c_idx=-1;
            break;

        default:
            bus_type="i2c";
            printf("Found chip: %s-%s-%d-%x\n", chip_name->prefix, bus_type,
                   i2c_nr, chip_name->addr);
            i2c_idx=i2c_nr;
            i2c_nr++;
            break;
        }

        fidx1=fidx2=0;

        /* Go into the sensor feature detection loop. */
        for (feature_id=sensors_get_all_features(*chip_name, &fidx1, &fidx2);
             feature_id!=NULL;
             feature_id=sensors_get_all_features(*chip_name, &fidx1, &fidx2)) {
            char *lab = NULL;

            /* Should this feature be skipped? */
            if (feature_id->mode!=SENSORS_MODE_R
             || feature_id->mapping!=SENSORS_NO_MAPPING
             || sensors_get_ignored(*chip_name, feature_id->number)<=0
             || sensors_get_label(*chip_name, feature_id->number, &lab)<0
             || lab==NULL) {
                continue;
            }

            printf("Found feature %d: %s\n", feature_id->number, lab);

            /* Stick this sensor in the list. */
            if (new_sensor(chip_name, lab, i2c_idx, feature_id->number)==NULL) {
                fprintf(stderr, "WARNING: allocating sensor data: %s\n",
                        strerror(errno));

                fprintf(stderr,
                        "WARNING: (did you forget to assign a label to %s?)\n",
                        lab);
            }

            free(lab);
        }
    }

    if (list_head==NULL) {
        /* We couldn't find any sensors? */
        fprintf(stderr, "FATAL: No sensors found!\n");
        goto fail_conf_open;
    }

    /* Now figure out how many polling intervals we'll go through... */
    total_msecs=minutes*60*1000;

    /* Now go into the loop. */
    printf("Going into scanning loop...");
    secs=total_msecs/1000;
    mins=secs/60;
    secs%=60;
    hours=mins/60;
    mins%=60;
    printf("%c[30GETA %d:%.2d:%.2d", 27, hours, mins, secs);
    fflush(stdout);

    for (cur_msecs=0; cur_msecs<total_msecs; ) {
        for (lsens=list_head; lsens!=NULL; lsens=lsens->next) {
            read_sensor(lsens);
        }

        mod=1000-(cur_msecs%1000);

        if (delay<mod) {
            /* First case: the next sensor checkpoint will come before the
             * next one-second milestone.
             */
            usleep(delay*1000);
            cur_msecs+=delay;
        }
        else {
            /* Second case: the next sensor checkpoint will come during or
             * after the next one-second milestone.
             */
            usleep(mod*1000);
            cur_msecs+=mod;
            secs=(total_msecs-cur_msecs)/1000;
            mins=secs/60;
            secs%=60;
            hours=mins/60;
            mins%=60;
            printf("%c[30GETA %d:%.2d:%.2d", 27, hours, mins, secs);
            fflush(stdout);

            for (mod=delay-mod; mod>=1000 && cur_msecs<total_msecs; mod-=1000) {
                /* There's at least another second before the next sensor
                 * checkpoint.
                 */
                usleep(1000000);
                cur_msecs+=1000;
                secs=(total_msecs-cur_msecs)/1000;
                mins=secs/60;
                secs%=60;
                hours=mins/60;
                mins%=60;
                printf("%c[30GETA %d:%.2d:%.2d", 27, hours, mins, secs);
                fflush(stdout);
            }

            /* The next sensor checkpoint is in less than a second... */
            if (cur_msecs<total_msecs) {
                usleep(mod*1000);
                cur_msecs+=mod;
            }
        }
    }

    /* Now write out the data we gathered. */
    printf("%c[28Gscanning complete.\n", 27);
    printf("Writing %s...", conf_file);

    if (write_header(conf_fd)<0) {
        fprintf(stderr, "FATAL: writing %s: %s\n", conf_file, strerror(errno));
        goto fail_list_alloced;
    }

    if (fprintf(conf_fd, "# This option specifies the polling interval in milliseconds.  It can only\n")<0
     || fprintf(conf_fd, "# be specified globally.\n")<0
     || fprintf(conf_fd, "DELAY %d\n\n", delay)<0) {
        fprintf(stderr, "FATAL: writing %s: %s\n", conf_file, strerror(errno));
        goto fail_list_alloced;
    }

    if (log_opt!=0) {
        if (fprintf(conf_fd, "# This option specifies whether or not to log sensor readings in non-alarm\n")<0
         || fprintf(conf_fd, "# states (alarm conditions are always logged).  It can be used either\n")<0
         || fprintf(conf_fd, "# globally or per-section.\n")<0
         || fprintf(conf_fd, "LOG 1\n\n")<0) {
            fprintf(stderr, "FATAL: writing %s: %s\n", conf_file, strerror(errno));
            goto fail_list_alloced;
        }
    }

    if (throttle>=0) {
        if (fprintf(conf_fd, "# This option instructs the daemon to throttle alarm repetition or log\n")<0
         || fprintf(conf_fd, "# repetition.  It can be used either globally or per-section.\n")<0
         || fprintf(conf_fd, "THROTTLE %d\n\n", throttle)<0) {
            fprintf(stderr, "FATAL: writing %s: %s\n", conf_file,
                    strerror(errno));

            goto fail_list_alloced;
        }
    }

    if (grace_period>0) {
        if (fprintf(conf_fd, "# !!! WARNING!  WARNING!  THIS IS A HACK! !!!\n")<0
         || fprintf(conf_fd, "# This option instructs the daemon to allow a grace period on executing alarm\n")<0
         || fprintf(conf_fd, "# handlers.  Effectively, a sensor has to go out of range %d times in a row\n", grace_period)<0
         || fprintf(conf_fd, "# before its alarm handler(s) will be executed.\n")<0
         || fprintf(conf_fd, "GRACE %d\n\n", grace_period)<0) {
            fprintf(stderr, "FATAL: writing %s: %s\n", conf_file,
                    strerror(errno));

            goto fail_list_alloced;
        }
    }

    for (lsens=list_head; lsens!=NULL; lsens=lsens->next) {
        if (lsens->i2c_num>=0 && lsens->i2c_num<10) {
            result=fprintf(conf_fd,
                           "\"%s-%s-%d-%x:%s\" {\n",
                           lsens->chip_id->prefix,
                           lsens->bus_type,
                           lsens->i2c_num,
                           lsens->chip_id->addr,
                           lsens->feature_label);
        }
        else {
            result=fprintf(conf_fd,
                           "\"%s-%s-%x:%s\" {\n",
                           lsens->chip_id->prefix,
                           lsens->bus_type,
                           lsens->chip_id->addr,
                           lsens->feature_label);
        }

        if (result<=0) {
            fprintf(stderr, "FATAL: writing %s: %s\n", conf_file, strerror(errno));
            goto fail_list_alloced;
        }

        result=fprintf(conf_fd, "    ABOVE %.2f \"halt -fp\"\n", lsens->upper);

        if (result<=0) {
            fprintf(stderr, "FATAL: writing %s: %s\n", conf_file, strerror(errno));
            goto fail_list_alloced;
        }

        result=fprintf(conf_fd, "    BELOW %.2f \"halt -fp\"\n", lsens->lower);

        if (result<=0) {
            fprintf(stderr, "FATAL: writing %s: %s\n", conf_file,
                    strerror(errno));

            goto fail_list_alloced;
        }

        result=fprintf(conf_fd, "}\n\n");

        if (result<=0) {
            fprintf(stderr, "FATAL: writing %s: %s\n", conf_file,
                    strerror(errno));

            goto fail_list_alloced;
        }
    }

    result=fprintf(conf_fd, "# End of sample mondo.conf\n");

    if (result<=0) {
        fprintf(stderr, "FATAL: writing %s: %s\n", conf_file, strerror(errno));
        goto fail_list_alloced;
    }

    printf("done.\n");
    printf("Be sure to edit this file yourself and make sure all values are suitable;\n");
    printf("the limits I have set are probably not as forgiving as you would like.\n\n");

    while (list_head!=NULL) {
        lsens=list_head;
        list_head=lsens->next;
        free(lsens->feature_label);
        free(lsens);
    }

    fclose(conf_fd);
    sensors_cleanup();
    fclose(sensors_fd);
    return 0;

    fail_list_alloced:
        while (list_head!=NULL) {
            lsens=list_head;
            list_head=lsens->next;
            free(lsens->feature_label);
            free(lsens);
        }

    fail_conf_open:
        fclose(conf_fd);

    fail_sensors_open:
        fclose(sensors_fd);
        printf("\n\n");
        return -1;
}

mondo_sensor *new_sensor(const sensors_chip_name *chip_id, const char *label, int i2c_nr, int feature)
{
    mondo_sensor *ret;

    if (chip_id==NULL || label==NULL) {
        errno=EINVAL;
        return NULL;
    }

    ret=(mondo_sensor *) malloc(sizeof(mondo_sensor));

    if (ret==NULL) {
        return NULL;
    }

    ret->feature_label=strdup(label);

    if (ret->feature_label==NULL) {
        goto fail_ret_alloced;
    }

    if(sensors_get_feature(*chip_id, feature, &(ret->upper))<0) {
        goto fail_label_alloced;
    }

    ret->lower=ret->upper;
    ret->chip_id=chip_id;
    ret->feature_id=feature;
    ret->i2c_num=i2c_nr;

    switch(chip_id->bus) {
    case SENSORS_CHIP_NAME_BUS_ISA:
        ret->bus_type="isa";
        break;

    default:
        ret->bus_type="i2c";
        break;
    }

    ret->next=NULL;

    /* Now add this one to the list. */
    if (list_head==NULL) {
        list_head=list_tail=ret;
    }
    else {
        list_tail->next=ret;
        list_tail=ret;
    }

    return ret;

    fail_label_alloced:
        free(ret->feature_label);

    fail_ret_alloced:
        free(ret);
        return NULL;
}

void read_sensor(mondo_sensor *sens)
{
    double val;

    if (sensors_get_feature(*(sens->chip_id), sens->feature_id, &val)==0) {
        /* We got a valid reading... */
        if (val>sens->upper) {
            /* This reading breaks the upper bound. */
            sens->upper=val;
        }

        if (val<sens->lower) {
            /* This reading breaks the lower bound. */
            sens->lower=val;
        }
    }
}

int write_header(FILE *fd)
{
    time_t cur_time;
    int result;

    result=fprintf(fd, "#\n# sample mondo.conf file\n");

    if (result<=0) {
        return -1;
    }

    time(&cur_time);
    result=fprintf(fd, "# created by mondo-setup on %s",
                   asctime(localtime(&cur_time)));

    if (result<=0) {
        return -1;
    }

    result=fprintf(fd, "#\n# This program is free software; you can redistribute it and/or modify\n");

    if (result<=0) {
        return -1;
    }

    result=fprintf(fd, "# it under the terms of version 2 of the GNU General Public License as\n");

    if (result<=0) {
        return -1;
    }

    result=fprintf(fd, "# published by the Free Software Foundation.\n#\n");

    if (result<=0) {
        return -1;
    }

    result=fprintf(fd, "# Be sure to modify the limits listed below according to taste.\n#\n\n");

    if (result<=0) {
        return -1;
    }

    return 0;
}

void usage(const char *cmd, FILE *fd)
{
    if (fd==NULL) {
        fd=stdout;
    }

    fprintf(fd,
"USAGE: %s [-fhlV] [-o config_file] [-d delay] [-t minutes]\n"
"          [-g grace_period] [-T throttle] [-s sensors_file]\n"
"\n"
"  -o config_file  : specify the file to write configuration info to\n"
"                    (default is ./mondo.conf)\n"
"  -d delay        : polling interval in milliseconds (default is 4000)\n"
"  -t minutes      : time to run in minutes (default is 120, max is %d)\n"
"  -T throttle     : throttle parameter for all sensors (default is none)\n"
"  -s sensors_file : specify the location of lm_sensors sensors.conf\n"
"                    (default is %s)\n"
"  -f              : overwrite new configuration file if it already exists\n"
"  -h              : print out help (this screen)\n"
"  -l              : enable verbose logging parameter\n"
"  -g grace_period : do not execute alarm handlers unless a sensor goes\n"
"                    out of range grace_period times in a row.\n"
"\n"
"                    *** This is an ugly hack; DO NOT use this unless you\n"
"                    *** know exactly what it does.\n"
"\n"
"  -V              : print out version\n\n",
            cmd, MAX_SCAN_MINUTES, SENSORS_CONF_FILE);
}
