/***************************************************************************
                mondo.h  -  the meeting place of all functions.
                             -------------------
    begin                : Wed Dec 12 2001
    copyright            : (C) 2001 by Kelledin
    email                : kelledin@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of version 2 of the GNU General Public License as  *
 *   published by the Free Software Foundation.                            *
 *                                                                         *
 ***************************************************************************/

#ifndef __MONDO_H__
#  define __MONDO_H__

#  include <stdio.h>
#  include <stdlib.h>
#  include <unistd.h>
#  include <errno.h>
#  include <string.h>
#  include <sensors/sensors.h>
#  include <sys/types.h>
#  include <sys/wait.h>
#  include <sys/stat.h>
#  include <sys/time.h>
#  include <fcntl.h>
#  include <syslog.h>
#  include <signal.h>

#  include "ktio.h"

#  ifndef MONDO_CONF_FILE
#    define MONDO_CONF_FILE "/etc/mondo.conf"
#  endif

#  ifndef SENSORS_CONF_FILE
#    define SENSORS_CONF_FILE "/etc/sensors.conf"
#  endif

#  ifndef MONDO_VERSION
#    define MONDO_VERSION "0.1"
#  endif

typedef struct __tag_mondo_action {

    double                      limit;
    char                       *syscmd;
    struct __tag_mondo_action  *next;

} mondo_action;

typedef struct __tag_mondo_sensor {

    char                       *chip_label;
    char                       *feature_label;
    sensors_chip_name           chip_id;
    int                         feature_id;
    int                         throttle;
    int                         ticks;
    int                         log_ticks;
    int                         log;
    int                         grace_ticks;
    int                         grace;
    mondo_action               *ubound;
    mondo_action               *lbound;
    struct __tag_mondo_sensor  *next;

} mondo_sensor;

extern char  *conf_file;
extern int    mondo_delay;
extern int    mondo_log;
extern int    dolock;

#  ifdef __cplusplus

extern "C" {

#  endif

int read_conf();

int init_lm_api(const char *fname);
/* Added 12/13/2001 by Kelledin

    DESCRIPTION:
        Initializes the lm_sensors API based on the lm_sensors config file
        referred to by fname.

    PRECONDITIONS:
        sensors_init() should not have been called already, and fname should
        point to a valid NULL-terminated ASCII character array.  Wide and
        multi-byte characters are not supported.

        The calling program should not call lm_sensors functions directly,
        especially sensors_init() and sensors_cleanup().  Communications with
        lm_sensors libraries should take place only through the functions
        provided by hw.c.

    POSTCONDITIONS:
        sensors_init(fname) will be called, and various other internal data
        structures will be initialized in hw.c.

    RETURNS:
        If successful, init_lm_api() will return 0.  Otherwise, init_lm_api()
        will return a nonzero value and set errno to indicate what croaked.

    PROVIDED BY:
        hw.c

*/

int shutdown_lm_api();
/* Added 12/13/2001 by Kelledin

    DESCRIPTION:
        Calls cleanup_sensors() and de-initializes the internal structures
        opened by init_lm_api().

    PRECONDITIONS:
        The lm_sensors API should already be initialized by init_lm_api().

    POSTCONDITIONS:
        The lm_sensors API and hw.c's internal structures will be
        de-initialized.  The global mondo_sensors list will be freed.

    RETURNS:
        If successful, shutdown_lm_api() returns 0.  Otherwise,
        shutdown_lm_api() returns a nonzero value and sets errno to indicate
        what croaked.

    PROVIDED BY:
        hw.c

*/

mondo_sensor *locate_sensor(const char *sensor_name);
/* Added 12/13/2001 by Kelledin

    DESCRIPTION:
        Attempts to find a certain sensor feature, initialize a
        mondo_sensor struct for that sensor feature, and add that struct
        to the global mondo_sensor list.  sensor_name describes an lm_sensors
        chip and feature via format "<chip_label>:<feature_label>".  For
        example, a valid sensor_name might be:

            via686a-isa-6000:3.3V

        where "via686a-isa-6000" would be a health monitor chip that
        lm_sensors knows about, and "3.3V" would be the label associated with
        the sensor on that chip which measures the system I/O voltage.

        sensor_name is case sensitive.

    PRECONDITIONS:
        The lm_sensors API should already be initialized via init_lm_api(),
        and sensor_name must be a valid ASCII NULL-terminated character
        array.  Wide and multi-byte character arrays are not supported.

    POSTCONDITIONS:
        If successful, locate_sensor() will malloc a mondo_sensor struct
        and set its members such that the sensor's status can be read quickly
        via read_sensor().  It will set the struct's ubound and lbound members
        to NULL, disable the throttle setting, and place it at the end of the
        global sensors list.

    RETURNS:
        If successful, locate_sensor() will return a pointer to the newly
        malloc'd mondo_sensor struct.  If unsuccessful, locate_sensor()
        will return NULL and set errno to indicate what croaked.  The
        following errno values signify the following conditions:

            EPERM   The lm_sensors API has not been properly initialized;
                    probably init_lm_api() has not been called.

            EINVAL  sensor_name is an invalid specifier; perhaps it does
                    not have exactly two fields.

            ENOMEM  There was not enough memory to initialize a new
                    mondo_sensor struct.

            ENODEV  The chip field of sensor_name refers to a chip that
                    lm_sensors has not detected.  This may simply mean that
                    lm_sensors has not been configured correctly.

            ENOSYS  The feature field of sensor_name refers to a feature
                    that lm_sensors is currently not implementing.  This may
                    simply mean that lm_sensors has not been configured
                    correctly.

    PROVIDED BY:
        hw.c

*/

double read_sensor(const mondo_sensor *lsens);
/* Added 12/13/2001 by Kelledin

    DESCRIPTION:
        Reads the current sensor value measured by *lsens.

    PRECONDITIONS:
        lsens must pont to a valid mondo_sensor struct returned by
        locate_sensor().  The lm_sensors API should already be initialized
        vi init_lm_api().

    POSTCONDITIONS:
        None.

    RETURNS:
        If successful, read_sensor() returns the current value measured by
        the specified sensor.  If unsuccessful, read_sensor() returns 0.0
        and sets errno to indicate what croaked.  Since 0.0 is a legal return
        value, it is advisable to set errno to zero in order to reliably
        detect whether the function has failed or not.

    PROVIDED BY:
        hw.c

*/

void set_throttle_option(int throttle);
/* Added 1/16/2002 by Kelledin

    DESCRIPTION:
        If any sensor structs in the global sensors list do not have their
        throttle option set to anything, set_throttle_option will set their
        throttle option to the value of throttle.

    PRECONDITIONS:
        None.

    POSTCONDITIONS:
        All non-configured throttle options in the global sensor list will be
        set to the supplied setting.

    RETURNS:
        Returns nothing.

    PROVIDED BY:
        hw.c

*/

void set_log_option(int log);
/* Added 1/6/2002 by Kelledin

    DESCRIPTION:
        If any sensor structs in the global sensors list do not have their log
        option set to anything, set_log_option will set their log option to
        the value of log.

    PRECONDITIONS:
        None.

    POSTCONDITIONS:
        All non-configured log options in the global sensor list will be set
        to the supplied setting.

    RETURNS:
        Returns nothing.

    PROVIDED BY:
        hw.c

*/

void set_grace_option(int grace);

void free_sensor_list();
/* Added 12/13/2001 by Kelledin

    DESCRIPTION:
        Frees the global mondo_sensor list.

    PRECONDITIONS:
        None.

    POSTCONDITIONS:
        The global mondo_sensor list will be freed from memory.

    RETURNS:
        Returns nothing.

    PROVIDED BY:
        hw.c

*/

void sigalrm_handler(int sig);
/* Added 12/13/2001 by Kelledin

    DESCRIPTION:
        Handles SIGALRM signals from the system, which are generated at
        regular intervals as specified in the config file.  Upon receiving a
        SIGALRM, sigalrm_handler() will check all sensors to see if any
        have gone past their threshold.  If a sensor exceeds its limits,
        sigalrm_handler() will log a message about it and attempt to execute
        the command associated with this condition.

    PROVIDED BY:
        hw.c

*/

void sighup_handler ( int sig );
/* Added 12/13/2001 by Kelledin

    DESCRIPTION:
        Handles SIGHUP signals from the system.  Upon receiving a SIGHUP,
        sighup_handler() will attempt to reload its config file.  If it
        fails in this, it will clean up and exit.

    PROVIDED BY:
        misc.c

*/

void sigterm_handler(int sig);
/* Added 12/13/2001 by Kelledin

    DESCRIPTION:
        Handles SIGTERM signals from the system.  Upon receiving a SIGTERM,
        sigterm_handler() will clean up and exit.

    PROVIDED BY:
        misc.c

*/

void sigsegv_handler(int sig);
/* Added 12/13/2001 by Kelledin

    DESCRIPTION:
        Handles SIGSEGV signals from the system, in case we segfault.  This
        shouldn't happen, but in case it does, sigsegv_handler will log some
        informative messages and exit.

    PROVIDED BY:
        misc.c

*/

int set_alrm_timer(int msecs);
/* Added 12/13/2001 by Kelledin

    DESCRIPTION:
        Sets the process's real-time timer to generate a SIGALRM at regular
        intervals of msecs milliseconds.

    PRECONDITIONS:
        It is advisable to have the SIGALRM handler in place first.

    POSTCONDITIONS:
        The process's real-time countdown timer will begin generating SIGALRMs
        every msecs milliseconds.

    RETURNS:
        If successful, set_alrm_timer() returns 0; otherwise, set_alrm_timer()
        returns -1 and sets errno to indicate what croaked.

    PROVIDED BY:
        misc.c

*/

int set_signal_handlers();
/* Added 12/13/2001 by Kelledin

    DESCRIPTION:
        Sets the signal handlers for SIGTERM, SIGHUP, SIGSEGV, and SIGALRM.

    PRECONDITIONS:
        None.

    POSTCONDITIONS:
        Our own custom handlers for the above signals will be installed.

    RETURNS:
        If successful, set_signal_handlers() returns 0; otherwise,
        set_signal_handlers() returns -1 and sets errno to indicate what
        croaked.

    PROVIDED BY:
        misc.c

*/

int lock_pidfile();
/* Added 12/13/2001 by Kelledin

    DESCRIPTION:
        Opens /var/run/mondo.pid, establishes a POSIX write lock, and writes
        the current process ID to the file.

    PRECONDITIONS:
        This should probably be called _after_ we daemonize, or in the case
        of multithreaded apps, after we spawn the first thread.  Daemonizing
        and spawning threads has a way of disrupting POSIX locks. =/

    POSTCONDITIONS:
        /var/run/mondo.pid will be opened and locked for writing, and the
        current process ID will be stored in that file.

    RETURNS:
        If successful, lock_pidfile() returns 0; otherwise, lock_pidfile()
        returns -1 and sets errno to indicate what croaked.

    PROVIDED BY:
        misc.c

*/

int unlock_pidfile();
/* Added 12/13/2001 by Kelledin

    DESCRIPTION:
        Undoes all the work done by lock_pidfile().

    PRECONDITIONS:
        /var/run/mondo.pid should currently be locked by the current process
        via a call to lock_pidfile().

    POSTCONDITIONS:
        /var/run/mondo.pid will be unlocked, closed, and unlinked.

    RETURNS:
        If successful, unlock_pidfile() returns 0; otherwise, unlock_pidfile()
        returns -1 and sets errno to indicate what croaked.

    PROVIDED BY:
        misc.c

*/

char **parse_line(const char *line, char *delim, int *argc);
/* Added 12/13/2001 by Kelledin

    DESCRIPTION:
        Takes a NULL-terminated line and parses it into a list of parameters
        using the following rulesets:

            Each argument of the line is separated by one or more delimiter
            characters; these characters are listed in delim.

            Arguments can be surrounded by quotation marks ('"' characters).
            Any delimiters inside a pair of quotation marks are treated as
            as ordinary characters, not delimiters.  When an argument
            surrounded by quotation marks is extracted, the surrounding
            quotation marks are not extracted with it.

            "Escape sequences" are two-character sequences beginning with a
            backslash character ('\').  The second character of an escape
            sequence may be any character at all, except for an ASCII NULL
            terminator character.  When a parameter containing an escape
            escape sequence is extracted, the escape sequence translates to
            to just the second character of the sequence.  If this second
            second character is a delimiter or a quote character, it is not
            not treated as such, but is treated instead as an ordinary
            character.  The escape sequence "\\" will translate to a single
            single backslash when its parameter is extracted.

            The line may begin or end with any number of delimiters.

        The following conditions (and possibly others) cause a line to be
        considered invalid:

            A closing quote of a argument, followed by a character which is
            neither a delimiter nor a NULL terminator.  For example, the
            following line would be invalid:

                foo "bar""ba z"

            (The following line would be the corrected equivalent:)

                foo "bar" "ba z"

            A quoted argument with a missing closing quote.  For example,
            the following line would be invalid:

                foo "bar" "baz

            (The following line would be the correct equivalent:)

                foo "bar" "ba z"

            A non-quoted argument with a non-escaped quotation mark in the
            middle.  For example, the following line would be invalid:

                foo ba"r baz

            (The following line would possibly be the correct equivalent:)

                foo ba\"r baz

    PRECONDITIONS:
        line must point to a valid ASCII NULL-terminated character array.
        Wide and multi-byte characters are not supported.

    POSTCONDITIONS:
        A new array of character pointers will be allocated via malloc(), and
        *argc will be set to the number of elements in this array.  Note that
        in addition to the first *argc elements of the array, there is one
        extra element at index [*argc] which is set to NULL.  This is for the
        benefit of functions like execvp().

        Each element of the array will be allocated via malloc() and set to
        an argument contained in line.  Arguments will be parsed according to
        the rules stated above.  For example, passing the following character
        array as line:

            foo ba\ \"r "b a\"z" blah "bl\\\ ah"

        would return with the arguments parsed into the following list:

            foo
            ba "r
            b a"z
            blah
            bl\ ah

        It is the responsibility of the calling program to call free() upon
        each element of the array, as well as the array itself, at some point
        before the program exits.

    RETURNS:
        On success, parse() returns a malloc'd array of malloc'd character
        pointers and sets *argc to the number of elements in this array.
        On failure, parse() returns NULL and sets errno to indicate what
        croaked.

    PROVIDED BY:
        misc.c

*/

int add_ubound(mondo_action **ulist, double lim, char *cmd);

int add_lbound(mondo_action **llist, double lim, char *cmd);

void free_alist(mondo_action *alist);

#  ifdef __cplusplus

}

#  endif

#endif
